/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package moe.nekojimi.chords;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.security.auth.login.LoginException;
import net.dv8tion.jda.api.JDA;
import net.dv8tion.jda.api.JDABuilder;
import net.dv8tion.jda.api.entities.*;
import net.dv8tion.jda.api.events.message.guild.GuildMessageReceivedEvent;
import net.dv8tion.jda.api.hooks.ListenerAdapter;
import net.dv8tion.jda.api.managers.AudioManager;
import net.dv8tion.jda.api.requests.GatewayIntent;
import net.dv8tion.jda.api.utils.Compression;
import net.dv8tion.jda.api.utils.cache.CacheFlag;

/**
 *
 * @author jimj316
 */
public class Main extends ListenerAdapter
{

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) throws LoginException
    {
        // We only need 2 gateway intents enabled for this example:
        EnumSet<GatewayIntent> intents = EnumSet.of(
                // We need messages in guilds to accept commands from users
                GatewayIntent.GUILD_MESSAGES,
                // We need voice states to connect to the voice channel
                GatewayIntent.GUILD_VOICE_STATES
        );

        JDABuilder builder = JDABuilder.createDefault(args[0], intents);

        // Disable parts of the cache
        builder.disableCache(CacheFlag.MEMBER_OVERRIDES, CacheFlag.VOICE_STATE);
        // Enable the bulk delete event
        builder.setBulkDeleteSplittingEnabled(false);
        // Disable compression (not recommended)
        builder.setCompression(Compression.NONE);
        // Set activity (like "playing Something")
        builder.setActivity(Activity.playing("music!"));

        builder.addEventListeners(new Main());

        JDA jda = builder.build();
    }

    @Override
    public void onGuildMessageReceived(GuildMessageReceivedEvent event)
    {
        Message message = event.getMessage();
        User author = message.getAuthor();
        String content = message.getContentRaw();
        Guild guild = event.getGuild();

        // Ignore message if bot
        if (author.isBot())
            return;

        try
        {

            String arg = "";
            if (content.contains(" "))
                arg = content.split(" ", 2)[1];

            if (content.startsWith("!join "))
                onJoinCommand(event, guild, arg);
            else if (content.equals("!join"))
                onJoinCommand(event);
            else if (content.startsWith("!play "))
                onPlayCommand(event, guild, arg);
            else if (content.startsWith("!queue"))
                onQueueCommand(event, guild);
            else if (content.startsWith("!remove "))
                onRemoveCommand(event, guild, arg);
            else if (content.startsWith("!restart"))
                onRestartCommand(event);
            else if (content.startsWith("!skip"))
                onSkipCommand(event);
            else if (content.startsWith("!"))
                onHelpCommand(event);
        } catch (Exception ex)
        {
            event.getChannel().sendMessage("Error in command! " + ex.getMessage()).queue();
        }
    }

    /**
     * Handle command without arguments.
     *
     * @param event
     * The event for this command
     */
    private void onJoinCommand(GuildMessageReceivedEvent event)
    {
        // Note: None of these can be null due to our configuration with the JDABuilder!
        Member member = event.getMember();                              // Member is the context of the user for the specific guild, containing voice state and roles
        GuildVoiceState voiceState = member.getVoiceState();            // Check the current voice state of the user
        VoiceChannel channel = voiceState.getChannel();                 // Use the channel the user is currently connected to
        if (channel != null)
        {
            connectTo(channel);                                         // Join the channel of the user
            onConnecting(channel, event.getChannel());                  // Tell the user about our success
        } else
            onUnknownChannel(event.getChannel(), "your voice channel"); // Tell the user about our failure
    }

    /**
     * Handle command with arguments.
     *
     * @param event
     * The event for this command
     * @param guild
     * The guild where its happening
     * @param arg
     * The input argument
     */
    private void onJoinCommand(GuildMessageReceivedEvent event, Guild guild, String arg)
    {
        boolean isNumber = arg.matches("\\d+"); // This is a regular expression that ensures the input consists of digits
        VoiceChannel channel = null;
        if (isNumber) // The input is an id?
            channel = guild.getVoiceChannelById(arg);
        if (channel == null)                    // Then the input must be a name?
        {
            List<VoiceChannel> channels = guild.getVoiceChannelsByName(arg, true);
            if (!channels.isEmpty())            // Make sure we found at least one exact match
                channel = channels.get(0);      // We found a channel! This cannot be null.
        }

        TextChannel textChannel = event.getChannel();
        if (channel == null)                    // I have no idea what you want mr user
        {
            onUnknownChannel(textChannel, arg); // Let the user know about our failure
            return;
        }
        connectTo(channel);                     // We found a channel to connect to!
        onConnecting(channel, textChannel);     // Let the user know, we were successful!
    }

    private void onPlayCommand(GuildMessageReceivedEvent event, Guild guild, String arg)
    {
        try
        {
            Song song = new Song(new URL(arg));
            event.getChannel().sendMessage("Downloading ...").queue();
            boolean ok = downloadSong(song);
            if (ok)
            {
                musicHandler.addSong(song);
                if (musicHandler.isPlaying())
                    musicHandler.setPlaying(true);
                event.getChannel().sendMessage("Downloaded and added to queue!").queue();
            }

        } catch (MalformedURLException ex)
        {
            event.getChannel().sendMessage("That's not a valid URL you idiot! " + ex.getMessage()).queue();
        } catch (IOException | InterruptedException | RuntimeException ex)
        {
            event.getChannel().sendMessage("Failed to download! Reason: " + ex.getMessage()).queue();
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    private void onRestartCommand(GuildMessageReceivedEvent event)
    {
        // TODO: this needs to clear the current data queue
        boolean ok = musicHandler.restartSong();
        if (ok)
            event.getChannel().sendMessage("Restarted current song!").queue();
        else
            event.getChannel().sendMessage("Cannot restart!").queue();
    }

    private void onSkipCommand(GuildMessageReceivedEvent event)
    {
        boolean ok = musicHandler.nextSong(true);
        if (ok)
            event.getChannel().sendMessage("Skipped to next song!").queue();
        else
            event.getChannel().sendMessage("There's no more songs!").queue();
    }

    private void onQueueCommand(GuildMessageReceivedEvent event, Guild guild)
    {
        Queue<Song> songQueue = musicHandler.getSongQueue();
        String message = new String();
        int i = 1;
        message += "Now playing: " + musicHandler.getCurrentSong() + "\n";
        message += "---\n";
        for (Song song : songQueue)
            message += (i) + ": " + song + "\n";
        event.getChannel().sendMessage(message).queue();
    }

    private void onRemoveCommand(GuildMessageReceivedEvent event, Guild guild, String arg)
    {
        try
        {
            int i = Integer.parseInt(arg);
            boolean removed = musicHandler.removeSong(i - 1);
            final int size = musicHandler.getSongQueue().size();
            if (removed)
                event.getChannel().sendMessage("Song removed.").queue();
            else if (size > 1)
                event.getChannel().sendMessage("That's not a number between 1 and " + size + "!").queue();
            else if (size == 1)
                event.getChannel().sendMessage("There's only one song to remove and that's not one of them!").queue();

        } catch (NumberFormatException ex)
        {
            event.getChannel().sendMessage(arg + " isn't a number!").queue();
        }
    }

    private void onHelpCommand(GuildMessageReceivedEvent event)
    {
        String help = "Commands available:\n"
                + "!join <Channel>       - Joins a voice channel\n"
                + "!play <URL>           - Downloads the track at that URL and adds it to the queue.\n"
                + "!queue                - Show the songs currently playing and the current queue.\n"
                + "!remove <Index>       - Remove the song at position <Index> from the queue.\n"
                + "!skip                 - Skip the current song and play the next one.\n"
                + "!restart              - Try playing the current song again in case it goes wrong.\n";
        event.getChannel().sendMessage(help).queue();
    }

    private boolean downloadSong(Song song) throws IOException, RuntimeException, InterruptedException
    {
        String cmd = "/usr/bin/youtube-dl -x --audio-format=wav --no-playlist --write-info-json " + song.getUrl().toString();
        System.out.println("Running command: " + cmd);
//		Process exec = Runtime.getRuntime().exec().split(" "));
        Process exec = new ProcessBuilder(cmd.split(" ")).redirectOutput(ProcessBuilder.Redirect.PIPE).start();
        exec.waitFor(10, TimeUnit.SECONDS);
        InputStream in = exec.getInputStream();
        String output = new String(in.readAllBytes(), Charset.defaultCharset());
        System.out.println(output);
        Matcher matcher = Pattern.compile("Destination: (.*\\.wav)").matcher(output);
        if (matcher.find())
            song.setLocation(new File(matcher.group(1)));
        else
            if (exec.exitValue() != 0)
                throw new RuntimeException("youtube-dl failed with error " + exec.exitValue());
        return true;
//        String destination = matcher.group(1);
//        return destination;
    }

    /**
     * Inform user about successful connection.
     *
     * @param channel
     * The voice channel we connected to
     * @param textChannel
     * The text channel to send the message in
     */
    private void onConnecting(VoiceChannel channel, TextChannel textChannel)
    {
        textChannel.sendMessage("Connecting to " + channel.getName()).queue(); // never forget to queue()!
    }

    /**
     * The channel to connect to is not known to us.
     *
     * @param channel
     * The message channel (text channel abstraction) to send failure
     * information to
     * @param comment
     * The information of this channel
     */
    private void onUnknownChannel(MessageChannel channel, String comment)
    {
        channel.sendMessage("Unable to connect to ``" + comment + "``, no such channel!").queue(); // never forget to queue()!
    }

    /**
     * Connect to requested channel and start echo handler
     *
     * @param channel
     * The channel to connect to
     */
    private void connectTo(VoiceChannel channel)
    {
        Guild guild = channel.getGuild();
        // Get an audio manager for this guild, this will be created upon first use for each guild
        AudioManager audioManager = guild.getAudioManager();
        musicHandler = new MusicHandler();
        // Create our Send/Receive handler for the audio connection
//        EchoHandler handler = new EchoHandler();

        // The order of the following instructions does not matter!
        // Set the sending handler to our echo system
        audioManager.setSendingHandler(musicHandler);
        // Set the receiving handler to the same echo system, otherwise we can't echo anything
//        audioManager.setReceivingHandler(handler);
        // Connect to the voice channel
        audioManager.openAudioConnection(channel);
    }
    private MusicHandler musicHandler;

}
